"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _configDir, _feedsConfigMemCache, _feedIdsByNameMemCache, _configDir_1, _feedsConfigManager, _localFeedsDatabase;
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const crypto_util_1 = require("../../common/crypto_util");
const GarbageMap_1 = __importDefault(require("../../common/GarbageMap"));
const core_1 = require("../../interfaces/core");
const LocalFeedsDatabase_1 = __importDefault(require("./LocalFeedsDatabase"));
const isFeedConfig = (x) => {
    return core_1._validateObject(x, {
        feedId: core_1.isFeedId,
        privateKey: core_1.isPrivateKeyHex
    });
};
class FeedsConfigManager {
    constructor(configDir, opts) {
        this.opts = opts;
        _configDir.set(this, void 0);
        _feedsConfigMemCache.set(this, new GarbageMap_1.default(core_1.scaledDurationMsec(60 * 60 * 1000)));
        _feedIdsByNameMemCache.set(this, new GarbageMap_1.default(core_1.scaledDurationMsec(60 * 60 * 1000)));
        __classPrivateFieldSet(this, _configDir, configDir);
    }
    async addFeed(feedId, privateKey) {
        const c = await this.getFeedConfig(feedId);
        if (c) {
            throw Error('Cannot add feed. Feed with this ID already exists');
        }
        const dirPath = await this._feedConfigDirectory(feedId, { create: true });
        const config = {
            feedId,
            privateKey
        };
        await writeJsonFile(`${dirPath}/config.json`, config);
        __classPrivateFieldGet(this, _feedsConfigMemCache).set(feedId, { feedId, privateKey });
    }
    async getFeedConfig(feedId) {
        const c = __classPrivateFieldGet(this, _feedsConfigMemCache).get(feedId);
        if ((c) && (this.opts.useMemoryCache)) {
            return c;
        }
        const dirPath = await this._feedConfigDirectory(feedId, { create: false });
        if (fs_1.default.existsSync(dirPath.toString())) {
            const config = await readJsonFile(`${dirPath}/config.json`, {});
            if (isFeedConfig(config)) {
                __classPrivateFieldGet(this, _feedsConfigMemCache).set(feedId, config);
                return config;
            }
            else {
                throw Error(`Problem with feed config file: ${dirPath}/config.json`);
            }
        }
        return null;
    }
    async deleteFeed(feedId) {
        if (__classPrivateFieldGet(this, _feedsConfigMemCache).has(feedId)) {
            __classPrivateFieldGet(this, _feedsConfigMemCache).delete(feedId);
        }
        const dirPath = await this._feedConfigDirectory(feedId, { create: false });
        if (fs_1.default.existsSync(dirPath.toString())) {
            await fs_1.default.promises.rmdir(dirPath.toString(), { recursive: true });
        }
    }
    async setFeedIdForName(feedName, feedId) {
        const path = await this._feedIdByNamePath(feedName, { create: true });
        await writeJsonFile(path.toString(), { feedName, feedId });
        __classPrivateFieldGet(this, _feedIdsByNameMemCache).set(feedName, feedId);
    }
    async getFeedIdForName(feedName) {
        const f = __classPrivateFieldGet(this, _feedIdsByNameMemCache).get(feedName);
        if ((f) && (this.opts.useMemoryCache))
            return f;
        const path = await this._feedIdByNamePath(feedName, { create: false });
        if (!fs_1.default.existsSync(path.toString()))
            return null;
        const a = await readJsonFile(path.toString(), {});
        if (!core_1.isJSONObject(a))
            return null;
        const feedId = a.feedId;
        if (!feedId)
            return null;
        if (core_1.isFeedId(feedId)) {
            __classPrivateFieldGet(this, _feedIdsByNameMemCache).set(feedName, feedId);
            return feedId;
        }
        else
            return null;
    }
    async _feedConfigDirectory(feedId, opts) {
        const f = feedId;
        const ret = `${__classPrivateFieldGet(this, _configDir)}/feeds/${f[0]}${f[1]}/${f[2]}${f[3]}/${f[4]}${f[5]}/${f}`;
        if (opts.create) {
            if (!fs_1.default.existsSync(ret)) {
                await fs_1.default.promises.mkdir(ret, { recursive: true });
            }
        }
        return core_1.localFilePath(ret);
    }
    async _feedIdByNamePath(feedName, opts) {
        const f = feedName;
        const p = `${__classPrivateFieldGet(this, _configDir)}/feeds/byName`;
        if (opts.create) {
            if (!fs_1.default.existsSync(p)) {
                await fs_1.default.promises.mkdir(p, { recursive: true });
            }
        }
        return core_1.localFilePath(`${p}/${feedName}.json`);
    }
}
_configDir = new WeakMap(), _feedsConfigMemCache = new WeakMap(), _feedIdsByNameMemCache = new WeakMap();
class LocalFeedManager {
    // #localFeedsDatabaseOld: LocalFeedsDatabaseOld
    constructor(storageDir, configDir) {
        this.storageDir = storageDir;
        this.configDir = configDir;
        _configDir_1.set(this, void 0);
        _feedsConfigManager.set(this, void 0);
        _localFeedsDatabase.set(this, void 0);
        if (!fs_1.default.existsSync(storageDir.toString())) {
            throw Error(`Storage directory does not exist: ${storageDir}`);
        }
        __classPrivateFieldSet(this, _configDir_1, configDir);
        __classPrivateFieldSet(this, _feedsConfigManager, new FeedsConfigManager(__classPrivateFieldGet(this, _configDir_1), { useMemoryCache: true })); // todo: need to test this for both false and true
        __classPrivateFieldSet(// todo: need to test this for both false and true
        this, _localFeedsDatabase, new LocalFeedsDatabase_1.default(core_1.localFilePath(storageDir + '/feeds.db'))
        // this.#localFeedsDatabaseOld = new LocalFeedsDatabaseOld(storageDir)
        );
        // this.#localFeedsDatabaseOld = new LocalFeedsDatabaseOld(storageDir)
    }
    async createFeed(feedName) {
        // Create a new writeable feed on this node and return the ID of the new feed
        // Generate the crypto keypair. The publicKey determines the new feed ID
        const { publicKey, privateKey } = crypto_util_1.createKeyPair();
        const feedId = crypto_util_1.publicKeyHexToFeedId(crypto_util_1.publicKeyToHex(publicKey));
        await __classPrivateFieldGet(this, _feedsConfigManager).addFeed(feedId, crypto_util_1.privateKeyToHex(privateKey));
        if (feedName !== null) {
            await __classPrivateFieldGet(this, _feedsConfigManager).setFeedIdForName(feedName, feedId);
        }
        await __classPrivateFieldGet(this, _localFeedsDatabase).addFeed(feedId);
        // Return the feed ID
        return feedId;
    }
    async deleteFeed(feedId) {
        await __classPrivateFieldGet(this, _localFeedsDatabase).deleteFeed(feedId);
        await __classPrivateFieldGet(this, _feedsConfigManager).deleteFeed(feedId);
    }
    async getFeedId(feedName) {
        // assert(typeof(feedName) === 'string');
        // Look up the feed ID for a particular feed name by consulting the config file
        const feedId = await __classPrivateFieldGet(this, _feedsConfigManager).getFeedIdForName(feedName);
        if (feedId) {
            const existsLocally = await this.feedExistsLocally(feedId);
            if (!existsLocally)
                return null;
        }
        return feedId;
    }
    async hasWriteableFeed(feedId) {
        // Check whether this node has a writeable feed.
        // We do this by ensuring that we have the associated private key and that the feed directory exists
        const privateKey = await this.getPrivateKeyForFeed(feedId);
        if (!privateKey)
            return false;
        // Returns true if we have the writeable feed
        return await this.feedExistsLocally(feedId);
    }
    async getPrivateKeyForFeed(feedId) {
        // Consult the config to get the private key associated with a particular feed ID
        const c = await __classPrivateFieldGet(this, _feedsConfigManager).getFeedConfig(feedId);
        return c ? crypto_util_1.hexToPrivateKey(c.privateKey) : null;
    }
    async feedExistsLocally(feedId) {
        return await __classPrivateFieldGet(this, _localFeedsDatabase).hasFeed(feedId);
    }
    async getSignedSubfeedMessages(feedId, subfeedHash) {
        // Read the messages file
        return await __classPrivateFieldGet(this, _localFeedsDatabase).getSignedSubfeedMessages(feedId, subfeedHash);
    }
    async getSubfeedAccessRules(feedId, subfeedHash) {
        return await __classPrivateFieldGet(this, _localFeedsDatabase).getSubfeedAccessRules(feedId, subfeedHash);
    }
    async appendSignedMessagesToSubfeed(feedId, subfeedHash, messages) {
        await __classPrivateFieldGet(this, _localFeedsDatabase).appendSignedMessagesToSubfeed(feedId, subfeedHash, messages);
    }
    async setSubfeedAccessRules(feedId, subfeedHash, accessRules) {
        await __classPrivateFieldGet(this, _localFeedsDatabase).setSubfeedAccessRules(feedId, subfeedHash, accessRules);
    }
}
exports.default = LocalFeedManager;
_configDir_1 = new WeakMap(), _feedsConfigManager = new WeakMap(), _localFeedsDatabase = new WeakMap();
const readJsonFile = async (path, defaultVal) => {
    try {
        const txt = await fs_1.default.promises.readFile(path, 'utf-8');
        return JSON.parse(txt);
    }
    catch (err) {
        return defaultVal;
    }
};
const writeJsonFile = async (path, obj) => {
    const txt = crypto_util_1.JSONStringifyDeterministic(obj, 4);
    await fs_1.default.promises.writeFile(path, txt);
};
//# sourceMappingURL=LocalFeedManager.js.map